// Copyright 2023 Georges Racinet <georges.racinet@octobus.net>
//
// This software may be used and distributed according to the terms of the
// GNU General Public License version 2 or any later version.
//
// SPDX-License-Identifier: GPL-2.0-or-later
//! Utilities for Mercurial content handling
use hg::changelog::Changelog;
use hg::errors::HgError;
use hg::manifest::{Manifest, ManifestEntry, Manifestlog};
use hg::repo::Repo;
use hg::revlog::{Node, NodePrefix, RevlogError, NULL_REVISION};
use hg::utils::hg_path::HgPath;

use crate::gitaly::ObjectType;

use super::git;
use super::oid::blob_oid;
use tracing::debug;

#[derive(Debug, Default)]
/// Common representation for metadata of content presented as if they were Git inner objects.
///
/// The intended application if for blob and trees gRPC methods. These metadata
/// are to be converted to final gRPC messages (including data chunks),
/// depending on actual needs.
///
/// Fields that are both not always needed and potentially expensive to compute are
/// enclosed in an [`Option`] and meant to be set by the service layer, since they
/// are specific to the method being implemented.
pub struct ObjectMetadata {
    pub oid: String,
    pub obj_type: ObjectType,
    pub size: i64,
    /// Object permissions, as expected by Gitaly clients, hence encoded as Git would.
    pub mode: i32,
    /// Useful in the `GetBlobs` gRPC method, but not in `GetBlob`.
    /// The first field is the revision (UTF-8 in both methods),
    /// the second is the path.
    pub revision_path: Option<(String, Vec<u8>)>,
}

/// Encode as Git would the file permissions recorded in a [`ManifestEntry`].
fn git_perms(entry: &ManifestEntry) -> Result<i32, HgError> {
    match entry.flags {
        None => Ok(git::OBJECT_MODE_NON_EXECUTABLE),
        Some(b'x') => Ok(git::OBJECT_MODE_EXECUTABLE),
        Some(b'l') => Ok(git::OBJECT_MODE_LINK),
        // 't' is actually only used in treemanifest, but stil possible
        Some(b't') => Ok(git::OBJECT_MODE_TREE),
        Some(f) => Err(HgError::CorruptedRepository(format!(
            "Unsupported manifest flag {}",
            f
        ))),
    }
}

/// Return data and metadata for an expected blob at given changeset and path.
///
/// The changeset is specified as a [`NodePrefix`], hence resolution can fail, with
/// [`RevlogError::InvalidRevision`] when it does not exist.
///
/// If the path does not exist in the specified changeset, `Ok(None)` is returned.
pub fn lookup_blob(
    repo: &Repo,
    changeset_node_prefix: NodePrefix,
    path: &[u8],
) -> Result<Option<(Vec<u8>, ObjectMetadata)>, RevlogError> {
    let changelog = repo.changelog()?;
    let manifestlog = repo.manifestlog()?;
    // Not using `repo.manifest_for_node()` because we will need the changelog entry
    // to provide a full changeset Node ID in the final construction.
    changelog_entry_manifest(&changelog, &manifestlog, changeset_node_prefix)?
        .map(|(cs_node, manifest)| {
            let hg_path = HgPath::new(path);
            manifest
                .find_by_path(hg_path)? // closure now has to return Result
                .map(|manifest_entry| read_blob(repo, &cs_node, manifest_entry, hg_path))
                .transpose() // to return Result
        })
        .transpose()
        .map(|oo| oo.flatten())
}

pub fn changelog_entry_manifest(
    changelog: &Changelog,
    manifestlog: &Manifestlog,
    changeset_node_prefix: NodePrefix,
) -> Result<Option<(Node, Manifest)>, RevlogError> {
    let rev = changelog.rev_from_node(changeset_node_prefix)?;
    if rev == NULL_REVISION {
        return Ok(None);
    }
    let changelog_entry = changelog.entry_for_rev(rev)?;
    let manifest = manifestlog.data_for_node(changelog_entry.data()?.manifest_node()?.into())?;
    Ok(Some((*changelog_entry.as_revlog_entry().node(), manifest)))
}

pub fn read_blob(
    repo: &Repo,
    changeset_node: &Node,
    manifest_entry: ManifestEntry,
    hg_path: &HgPath,
) -> Result<(Vec<u8>, ObjectMetadata), RevlogError> {
    {
        let file_node = manifest_entry.node_id()?;
        let file_log = repo.filelog(hg_path)?;
        let fl_data = file_log.data_for_node(file_node)?;
        let data = fl_data.into_file_data()?;
        let size = data.len();
        Ok((
            data,
            ObjectMetadata {
                size: size as i64,
                mode: git_perms(&manifest_entry)?,
                obj_type: ObjectType::Blob,
                oid: blob_oid(changeset_node, hg_path.as_bytes()),
                ..Default::default()
            },
        ))
    }
}

/// Relevant metadata a directory item , which can be a file or a sub-directory
///
/// The `manifest_entry` field is `None` if and only if the item is a sub-directory.
/// Otherwise the [`ManifestEntry`] provides metadata of the file.
#[derive(Debug)]
pub struct DirectoryEntry<'m> {
    pub relative_path: Vec<u8>,
    pub mode: i32,
    pub manifest_entry: Option<ManifestEntry<'m>>,
}

impl DirectoryEntry<'_> {
    pub fn is_file(&self) -> bool {
        self.manifest_entry.is_some()
    }
}

impl PartialEq for DirectoryEntry<'_> {
    fn eq(&self, other: &Self) -> bool {
        if self.relative_path != other.relative_path || self.mode != other.mode {
            false
        } else {
            match (self.manifest_entry.as_ref(), other.manifest_entry.as_ref()) {
                (Some(e), Some(f)) => manifest_entry_eq(e, f),
                (None, None) => true,
                _ => false,
            }
        }
    }
}

fn manifest_entry_eq<'m>(e: &ManifestEntry<'m>, f: &ManifestEntry<'m>) -> bool {
    // TODO upstream implement PartialEq for ManifestEntry, pretty much as this
    e.path == f.path && e.hex_node_id == f.hex_node_id
}

impl Eq for DirectoryEntry<'_> {}

/// Represent the content at some path, either a file, with a [`ManifestEntry`] for metadata,
/// or a directory with its entries.
#[derive(Debug)]
pub enum PathContent<'m> {
    File(ManifestEntry<'m>),
    Directory(Vec<DirectoryEntry<'m>>),
    NotFound,
}

impl PartialEq for PathContent<'_> {
    fn eq(&self, other: &Self) -> bool {
        match (self, other) {
            (PathContent::NotFound, PathContent::NotFound) => true,
            (PathContent::File(e), PathContent::File(f)) => manifest_entry_eq(e, f),
            (PathContent::Directory(v), PathContent::Directory(w)) => v == w,
            _ => false,
        }
    }
}

impl Eq for PathContent<'_> {}

/// Return the list of entries that are sitting right at path.
///
/// Similarly to `/bin/ls`, if `path` is a directory, its contents are listed, otherwise
/// file metadata are returned.
pub fn ls_path<'m>(manifest: &'m Manifest, path: &[u8]) -> Result<PathContent<'m>, RevlogError> {
    inner_ls_path(manifest.iter(), path)
}

fn inner_ls_path<'m, IM>(manifest: IM, path: &[u8]) -> Result<PathContent<'m>, RevlogError>
where
    IM: IntoIterator<Item = Result<ManifestEntry<'m>, HgError>>,
{
    let mut listing = Vec::new();
    let pl = path.len();

    for entry_res in manifest {
        let entry = entry_res?;
        let ep = entry.path.as_bytes();
        if !ep.starts_with(path) {
            if !listing.is_empty() {
                break;
            } else {
                continue;
            }
        }
        debug!("Considering manifest path {:?}", ep);
        // exact match is always first, because manifest is lexicographically ordered
        if ep.len() == pl {
            // so ep == path
            debug!("Exact match for {:?} and {:?}", ep, pl);
            return Ok(PathContent::File(entry));
        }

        debug!("Checking if we are in a subdir with {:?}", ep);
        if ep[pl] != b'/' {
            if !listing.is_empty() {
                break;
            } else {
                continue;
            }
        }

        let rel_path = &ep[pl + 1..];
        debug!("Analyzing subdir for {:?} (rel_path={:?})", ep, rel_path);
        // first iteration on split can be `None` only if rel_path is empty,
        // which would mean in this context that entry.path has a trailing slash
        // (guaranteed not to happen in the Manifest data structure)
        if let Some(sub_rel_path) = rel_path.split(|c| *c == b'/').next() {
            if sub_rel_path.len() == rel_path.len() {
                listing.push(DirectoryEntry {
                    relative_path: rel_path.to_vec(),
                    mode: git_perms(&entry)?,
                    manifest_entry: Some(entry),
                });
            } else {
                if let Some(previous) = listing.last() {
                    if previous.relative_path == sub_rel_path {
                        continue;
                    }
                }
                listing.push(DirectoryEntry {
                    relative_path: sub_rel_path.to_vec(),
                    mode: git::OBJECT_MODE_TREE,
                    manifest_entry: None,
                });
            }
        }
    }
    Ok(if listing.is_empty() {
        PathContent::NotFound
    } else {
        PathContent::Directory(listing)
    })
}

#[cfg(test)]
mod tests {

    use super::*;

    /// Return a [`ManifestEntry`] suitable for tests that care only about paths
    fn path_manifest_entry(path: &'static str) -> ManifestEntry {
        ManifestEntry {
            path: HgPath::new(path.as_bytes()),
            hex_node_id: b"",
            flags: None,
        }
    }

    /// Return something suitable to test the algorithm behind [`ls_path`] from only paths.
    ///
    /// For [`ls_path`] correctness, the node ids and flags don't matter much, hence we give
    /// them fixed, arbitrary values.
    /// For convenience paths are string slices, static not to bother with lifetimes.
    fn paths_manifest(paths: Vec<&'static str>) -> Vec<Result<ManifestEntry, HgError>> {
        paths
            .into_iter()
            .map(|path| Ok(path_manifest_entry(path)))
            .collect()
    }

    #[test]
    fn test_ls_path_simple() {
        let manifest = paths_manifest(vec!["bar", "foo/some", "foo0"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"foo").unwrap(),
            PathContent::Directory(vec![DirectoryEntry {
                relative_path: b"some".to_vec(),
                mode: git::OBJECT_MODE_NON_EXECUTABLE,
                manifest_entry: Some(path_manifest_entry("foo/some")),
            }])
        );

        // other climbing case
        let manifest = paths_manifest(vec!["foo/some", "other"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"foo").unwrap(),
            PathContent::Directory(vec![DirectoryEntry {
                relative_path: b"some".to_vec(),
                mode: git::OBJECT_MODE_NON_EXECUTABLE,
                manifest_entry: Some(path_manifest_entry("foo/some")),
            }])
        );

        // other diving case
        let manifest = paths_manifest(vec!["foo.", "foo/some"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"foo").unwrap(),
            PathContent::Directory(vec![DirectoryEntry {
                relative_path: b"some".to_vec(),
                mode: git::OBJECT_MODE_NON_EXECUTABLE,
                manifest_entry: Some(path_manifest_entry("foo/some")),
            }])
        );

        let manifest = paths_manifest(vec!["bar", "foo/some", "foo0"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"bar").unwrap(),
            PathContent::File(path_manifest_entry("bar")),
        );

        let manifest = paths_manifest(vec!["bar", "foo/some", "foo0"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"foo0").unwrap(),
            PathContent::File(path_manifest_entry("foo0")),
        );

        let manifest = paths_manifest(vec!["bar", "foo"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"alien").unwrap(),
            PathContent::NotFound,
        );
    }

    #[test]
    fn test_ls_path_dir_several_files() {
        let manifest = paths_manifest(vec!["foo.", "foo/a", "foo/b", "foo0"]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"foo").unwrap(),
            PathContent::Directory(vec![
                DirectoryEntry {
                    relative_path: b"a".to_vec(),
                    mode: git::OBJECT_MODE_NON_EXECUTABLE,
                    manifest_entry: Some(path_manifest_entry("foo/a")),
                },
                DirectoryEntry {
                    relative_path: b"b".to_vec(),
                    mode: git::OBJECT_MODE_NON_EXECUTABLE,
                    manifest_entry: Some(path_manifest_entry("foo/b")),
                }
            ])
        )
    }
    #[test]
    fn test_ls_path_dir_with_sub() {
        let manifest = paths_manifest(vec![
            "foo.",
            "foo/a",
            "foo/sub/a",
            "foo/sub/b",
            "foo/subb",
            "foo0",
        ]);
        assert_eq!(
            inner_ls_path(manifest.into_iter(), b"foo").unwrap(),
            PathContent::Directory(vec![
                DirectoryEntry {
                    relative_path: b"a".to_vec(),
                    mode: git::OBJECT_MODE_NON_EXECUTABLE,
                    manifest_entry: Some(path_manifest_entry("foo/a")),
                },
                DirectoryEntry {
                    relative_path: b"sub".to_vec(),
                    mode: git::OBJECT_MODE_TREE,
                    manifest_entry: None,
                },
                DirectoryEntry {
                    relative_path: b"subb".to_vec(),
                    mode: git::OBJECT_MODE_NON_EXECUTABLE,
                    manifest_entry: Some(path_manifest_entry("foo/subb")),
                },
            ])
        )
    }
}
