use std::fmt::{Debug, Formatter};
use std::sync::Arc;
use tokio_stream::wrappers::ReceiverStream;
use tonic::{metadata::Ascii, metadata::MetadataValue, Request, Response, Status};
use tracing::{field, info, instrument};

use hg::changelog::Changelog;
use hg::repo::Repo;
use hg::revlog::{NodePrefix, RevlogError, NULL_REVISION};
use hg::utils::hg_path::HgPath;

use super::blob::stream_blob;
use crate::config::Config;
use crate::git;
use crate::gitaly::commit_service_server::{CommitService, CommitServiceServer};
use crate::gitaly::{
    tree_entry_response::ObjectType, FindCommitRequest, FindCommitResponse,
    ListCommitsByOidRequest, ListCommitsByOidResponse, Repository, TreeEntryRequest,
    TreeEntryResponse,
};
use crate::gitlab::revision::gitlab_revision_node_prefix;
use crate::mercurial::{changelog_entry_manifest, ls_path, read_blob, PathContent};
use crate::message;
use crate::metadata::correlation_id;
use crate::oid::tree_oid;
use crate::repository::{
    default_repo_spec_error_status, load_repo_and_stream, load_repo_and_then, repo_store_vfs,
    RequestWithRepo,
};
use crate::streaming::{BlockingResponseSender, ResultResponseStream};

#[derive(Debug)]
pub struct CommitServiceImpl {
    config: Arc<Config>,
}

const COMMITS_CHUNK_SIZE: usize = 50;

/// Return the [`ListCommitsByOidResponse`] for a chunk of oids (changeset ids)
///
/// Behaving like Gitaly imposes that
/// - oids that cannot be found in the repository are ignored. This is actually one of the
///   use cases for the `ListCommitsByOid` call: it can be called indiscriminately with anything
///   that looks like a hash, in particular from Markdown renderings
/// - oids that are not even hexadecimal are ignored, too. This is possibly coincidental.
///
/// [`RevlogError`] is returned in cases of repository corruption.
fn commits_by_oid_chunk(
    cl: &Changelog,
    oids: &[String],
) -> Result<ListCommitsByOidResponse, RevlogError> {
    // Not using Vec::with_capacity(oids.len()), because clients will often call this with a single
    // argument, precisely to know whether it is a commit id prefix in this repo (happens a lot
    // in Markdown rendering), and this will result in needless allocations.
    let mut commits = Vec::new();
    for oid in oids {
        match NodePrefix::from_hex(oid) {
            Err(_) => {
                continue; // non-hexadecimal values for oid are just ignored by Gitaly
            }
            Ok(node_pref) => match cl.rev_from_node(node_pref) {
                Ok(NULL_REVISION) => (),
                Ok(rev) => commits.push(message::commit(cl.entry_for_rev(rev)?)?),
                Err(RevlogError::InvalidRevision) => (),
                Err(e) => {
                    return Err(e);
                }
            },
        }
    }
    Ok(ListCommitsByOidResponse { commits })
}

fn not_found_for_path(path: &[u8]) -> Status {
    Status::not_found(format!("not found: {}", &String::from_utf8_lossy(path)))
}

fn stream_tree_entry(
    req: TreeEntryRequest,
    repo: &Repo,
    node_prefix: NodePrefix,
    tx: &BlockingResponseSender<TreeEntryResponse>,
) -> Result<(), Status> {
    let changelog = repo
        .changelog()
        .map_err(|e| Status::internal(format!("Error getting changelog: {:?}", e)))?;
    let manifestlog = repo
        .manifestlog()
        .map_err(|e| Status::internal(format!("Error getting manifestlog: {:?}", e)))?;
    let (cs_node, manifest) = changelog_entry_manifest(&changelog, &manifestlog, node_prefix)
        .map_err(|e| Status::internal(format!("Error obtaining manifest: {:?}", e)))?
        .ok_or_else(|| not_found_for_path(&req.path))?;

    let mut pl = req.path.len();
    while pl > 0 && req.path[pl - 1] == b'/' {
        pl -= 1;
    }
    let trimmed = &req.path[0..pl];

    match ls_path(&manifest, trimmed)
        .map_err(|e| Status::internal(format!("Error reading manifest {:?}", e)))?
    {
        PathContent::File(manifest_entry) => {
            let hg_path = HgPath::new(&req.path);
            let (mut data, metadata) = read_blob(repo, &cs_node, manifest_entry, hg_path)
                .map_err(|e| Status::internal(format!("Error reading file {:?}", e)))?;
            if req.max_size != 0 && metadata.size > req.max_size {
                tx.send(Err(Status::failed_precondition(format!(
                    "TreeEntry: object size ({}) is bigger than the maximum \
                                         allowed size ({})",
                    metadata.size, req.max_size
                ))));
            } else {
                if req.limit > 0 {
                    data.truncate(req.limit as usize);
                }
                stream_blob(tx, data, metadata)
            }
        }
        PathContent::NotFound => {
            return Err(not_found_for_path(&req.path));
        }
        PathContent::Directory(listing) => {
            let mut size: usize = 0;
            // See Python impl for explanation (note that the Rust implementation is using
            // the relative path instead of the absolute path).
            for elt in listing.iter() {
                if elt.is_file() {
                    size += 28
                } else {
                    size += 27
                }
                size += elt.relative_path.len();
            }
            tx.send(Ok(TreeEntryResponse {
                r#type: ObjectType::Tree as i32,
                oid: tree_oid(&cs_node, trimmed),
                mode: git::OBJECT_MODE_TREE,
                size: size as i64,
                ..Default::default()
            }));
        }
    }
    Ok(())
}

impl RequestWithRepo for TreeEntryRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for ListCommitsByOidRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

impl RequestWithRepo for FindCommitRequest {
    fn repository_ref(&self) -> Option<&Repository> {
        self.repository.as_ref()
    }
}

#[tonic::async_trait]
impl CommitService for CommitServiceImpl {
    type ListCommitsByOidStream = ReceiverStream<Result<ListCommitsByOidResponse, Status>>;
    type TreeEntryStream = ReceiverStream<Result<TreeEntryResponse, Status>>;

    async fn list_commits_by_oid(
        &self,
        request: Request<ListCommitsByOidRequest>,
    ) -> Result<Response<Self::ListCommitsByOidStream>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_list_commits_by_oid(inner, correlation_id(&metadata))
            .await
    }

    async fn find_commit(
        &self,
        request: Request<FindCommitRequest>,
    ) -> Result<Response<FindCommitResponse>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_find_commit(inner, correlation_id(&metadata))
            .await
    }

    async fn tree_entry(
        &self,
        request: Request<TreeEntryRequest>,
    ) -> Result<Response<Self::TreeEntryStream>, Status> {
        let (metadata, _ext, inner) = request.into_parts();

        self.inner_tree_entry(inner, correlation_id(&metadata))
            .await
    }
}

struct TreeEntryTracingRequest<'a>(&'a TreeEntryRequest);

impl Debug for TreeEntryTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("TreeEntryRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .field("path", &String::from_utf8_lossy(&self.0.path))
            .field("limit", &self.0.limit)
            .field("max_size", &self.0.max_size)
            .finish()
    }
}

struct FindCommitTracingRequest<'a>(&'a FindCommitRequest);

impl Debug for FindCommitTracingRequest<'_> {
    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
        f.debug_struct("FindCommitRequest")
            .field("repository", &self.0.repository)
            .field("revision", &String::from_utf8_lossy(&self.0.revision))
            .finish()
    }
}

impl CommitServiceImpl {
    #[instrument(name = "tree_entry",
                 skip(self, request),
                 fields(request=field::debug(TreeEntryTracingRequest(&request))))]
    async fn inner_tree_entry(
        &self,
        request: TreeEntryRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<TreeEntryResponse> {
        info!("Processing");

        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        match gitlab_revision_node_prefix(&store_vfs, &request.revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => {
                info!("Revision not resolved");
                Err(not_found_for_path(&request.path))
            }
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);

                load_repo_and_stream(
                    self.config.clone(),
                    request,
                    default_repo_spec_error_status,
                    move |req, repo, tx| {
                        if let Err(e) = stream_tree_entry(req, &repo, node_prefix, &tx) {
                            tx.send(Err(e));
                        }
                    },
                )
            }
        }
    }

    #[instrument(name = "list_commits_by_oid", skip(self))]
    async fn inner_list_commits_by_oid(
        &self,
        request: ListCommitsByOidRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> ResultResponseStream<ListCommitsByOidResponse> {
        info!("Processing");

        load_repo_and_stream(
            self.config.clone(),
            request,
            default_repo_spec_error_status,
            move |req, repo, tx| match repo.changelog() {
                Ok(cl) => {
                    for oid_chunk in req.oid.chunks(COMMITS_CHUNK_SIZE) {
                        match commits_by_oid_chunk(&cl, oid_chunk) {
                            Ok(resp) => {
                                if !resp.commits.is_empty() {
                                    tx.send(Ok(resp));
                                }
                            }
                            Err(e) => tx.send(Err(Status::internal(format!(
                                "Repository corruption {:?}",
                                e
                            )))),
                        }
                    }
                }
                Err(e) => tx.send(Err(Status::internal(format!(
                    "Could not open repository changelog: {:?}",
                    e
                )))),
            },
        )
    }

    #[instrument(name="find_commit",
                 skip(self, request),
                 fields(request=field::debug(FindCommitTracingRequest(&request))))]
    async fn inner_find_commit(
        &self,
        request: FindCommitRequest,
        correlation_id: Option<&MetadataValue<Ascii>>,
    ) -> Result<Response<FindCommitResponse>, Status> {
        info!("Processing");
        if request.revision.is_empty() {
            return Err(Status::invalid_argument("empty revision"));
        }

        let config = self.config.clone();
        let store_vfs = repo_store_vfs(&config, &request.repository)
            .await
            .map_err(default_repo_spec_error_status)?;

        match gitlab_revision_node_prefix(&store_vfs, &request.revision)
            .await
            .map_err(|e| Status::internal(format!("Error resolving revision: {:?}", e)))?
        {
            None => {
                info!("Revision not resolved");
                Ok(Response::new(FindCommitResponse::default()))
            }
            Some(node_prefix) => {
                info!("Revision resolved as {:x}", &node_prefix);
                let commit = load_repo_and_then(
                    self.config.clone(),
                    request,
                    default_repo_spec_error_status,
                    move |_req, repo| {
                        let cl = repo.changelog().map_err(|e| {
                            Status::internal(format!("Could not open changelog: {:?}", e))
                        })?;

                        match message::commit_for_node_prefix(&cl, node_prefix) {
                            // TODO discuss upstream: InvalidRevision is incorrect wording:
                            // it does represent both invalid input (working dir Node),
                            // and not found node prefixes. The latter case justifies our
                            // choice.
                            // Note that at this stage we cannot tell apart NodePrefix due to
                            // caller passing a hash from those read in state files (would be an
                            // inconsistency for them not to resolve).
                            Err(RevlogError::InvalidRevision) => Ok(None),
                            Err(e) => {
                                Err(Status::internal(format!("Repository corruption {:?}", e)))
                            }
                            Ok(r) => Ok(r),
                        }
                    },
                )
                .await?;

                Ok(Response::new(FindCommitResponse { commit }))
            }
        }
    }
}

/// Takes care of boilerplate that would instead be in the startup sequence.
pub fn commit_server(config: &Arc<Config>) -> CommitServiceServer<CommitServiceImpl> {
    CommitServiceServer::new(CommitServiceImpl {
        config: config.clone(),
    })
}
